/* Copyright (c) 2006-2007 Christopher J. W. Lloyd

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. */
#include <math.h>

#import <Foundation/NSKeyedUnarchiver.h>
#import <Foundation/NSPropertyListReader.h>
#import <Foundation/NSString.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSNumber.h>
#import <Foundation/NSData.h>
#import <Foundation/NSException.h>
#import <Foundation/CFUID.h>
#import <Foundation/NSCFTypeID.h>

NSString* NSInvalidUnarchiveOperationException=@"NSInvalidUnarchiveOperationException";

@interface NSObject(NSKeyedUnarchiverPrivate)
+(id)allocWithKeyedUnarchiver:(NSKeyedUnarchiver *)keyed;
@end

@implementation NSObject(NSKeyedUnarchiverPrivate)

+(id)allocWithKeyedUnarchiver:(NSKeyedUnarchiver *)keyed {
   return [self allocWithZone:NULL];
}

@end

@implementation NSKeyedUnarchiver

-initForReadingWithData:(NSData *)data {
   _nameToReplacementClass=[NSMutableDictionary new];
   _propertyList=[[NSPropertyListReader propertyListFromData:data] retain];
   _objects=[[_propertyList objectForKey:@"$objects"] retain];
   _plistStack=[NSMutableArray new];
   [_plistStack addObject:[_propertyList objectForKey:@"$top"]];
   _uidToObject=NSCreateMapTable(NSIntMapKeyCallBacks,NSNonRetainedObjectMapValueCallBacks,0);
   return self;
}

-(void)dealloc {
   [_nameToReplacementClass release];
   [_propertyList release];
   [_objects release];
   [_plistStack release];
   if(_uidToObject!=NULL)
    NSFreeMapTable(_uidToObject);
   [super dealloc];
}

-(BOOL)allowsKeyedCoding {
   return YES;
}

static inline int integerFromCFUID(id object){
// We can deal with CFUID's and dictionaries
   unsigned typeID=[object _cfTypeID];
   
   if(typeID==kNSCFTypeDictionary){
    NSNumber *uid=[object objectForKey:@"CF$UID"];
    return [uid integerValue];
   }

   return [object integerValue];
}

-(Class)decodeClassFromDictionary:(NSDictionary *)classReference {
   Class         result;
   NSDictionary *plist=[classReference objectForKey:@"$class"];
   NSDictionary *profile=[_objects objectAtIndex:integerFromCFUID(plist)];
   NSDictionary *classes=[profile objectForKey:@"$classes"];
   NSString     *className=[profile objectForKey:@"$classname"];
   
   if((result=[_nameToReplacementClass objectForKey:className])==Nil)
    if((result=NSClassFromString(className))==Nil)
     [NSException raise:NSInvalidArgumentException format:@"Unable to find class named %@",className];
    
   return result;
}

-decodeObjectForUID:(NSInteger )uidIntValue {
   id result=NSMapGet(_uidToObject,(void *)uidIntValue);
            
   if(result==nil){
    id plist=[_objects objectAtIndex:uidIntValue];
    unsigned typeID=[plist _cfTypeID];
    
    if(typeID==kNSCFTypeString){
     if([plist isEqualToString:@"$null"])
      result=nil;
     else {
      result=plist;
      NSMapInsert(_uidToObject,(void *)uidIntValue,result);
     }
    }
    else if(typeID==kNSCFTypeDictionary){
     Class class=[self decodeClassFromDictionary:plist];
   
     [_plistStack addObject:plist];
     result=[class allocWithKeyedUnarchiver:self];
     NSMapInsert(_uidToObject,(void *)uidIntValue,result);
     result=[result initWithCoder:self];
     NSMapInsert(_uidToObject,(void *)uidIntValue,result);
     result=[result awakeAfterUsingCoder:self];
     [result autorelease];
     if([_delegate respondsToSelector:@selector(unarchiver:didDecodeObject:)])
      result=[_delegate unarchiver:self didDecodeObject:result];
     NSMapInsert(_uidToObject,(void *)uidIntValue,result);
     [_plistStack removeLastObject];
    }
    else if(typeID==kNSCFTypeNumber || typeID==kNSCFTypeBoolean){
     result=plist;
     NSMapInsert(_uidToObject,(void *)uidIntValue,result);
    }
    else if (typeID==kNSCFTypeData) {
     result=plist;
     NSMapInsert(_uidToObject,(void *)uidIntValue,result);
    }
    else
     NSLog(@"plist of class %@",[plist class]);
   }
   
   return result;  
}

-decodeRootObject {
   NSDictionary *top=[_propertyList objectForKey:@"$top"];
   NSArray      *values=[top allValues];

   if([values count]!=1){
    NSLog(@"multiple values=%@",values);
    return nil;
   }
   else {
    NSDictionary *object=[values objectAtIndex:0];
    
    return [self decodeObjectForUID:integerFromCFUID(object)];
   }
}

+unarchiveObjectWithData:(NSData *)data {
   NSKeyedUnarchiver *unarchiver=[[[self alloc] initForReadingWithData:data] autorelease];
   
   return [unarchiver decodeRootObject];
}

+unarchiveObjectWithFile:(NSString *)path {
   NSData *data=[NSData dataWithContentsOfFile:path];
   
   return [self unarchiveObjectWithData:data];
}

-(BOOL)containsValueForKey:(NSString *)key {
   return ([[_plistStack lastObject] objectForKey:key]!=nil)?YES:NO;
}

-(const uint8_t *)decodeBytesForKey:(NSString *)key returnedLength:(NSUInteger *)lengthp {
   NSData *data=[[_plistStack lastObject] objectForKey:key];

   *lengthp=[data length];
   
   return [data bytes];
}

static inline NSNumber *_numberForKey(NSKeyedUnarchiver *self,NSString *key){
   NSNumber *result=[[self->_plistStack lastObject] objectForKey:key];

   if(result==nil)
    return result;
    
   unsigned typeID=[result _cfTypeID];
  
   if(typeID==kNSCFTypeNumber || typeID==kNSCFTypeBoolean)
    return result;
    
   [NSException raise:@"NSKeyedUnarchiverException" format:@"Expecting number, got %@",result];
   return nil;
}

-(BOOL)decodeBoolForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return NO;
    
   return [number boolValue];
}

-(double)decodeDoubleForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return 0;

   return [number doubleValue];
}

-(float)decodeFloatForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return 0;
   
   return [number floatValue];
}

-(int)decodeIntForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return 0;
    
   return [number intValue];
}

-(int32_t)decodeInt32ForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return 0;
    
   return [number intValue];
}

-(int64_t)decodeInt64ForKey:(NSString *)key {
   NSNumber *number=_numberForKey(self,key);
   
   if(number==nil)
    return 0;
    
   return [number intValue];
}

// not a lot of validation
-(NSUInteger)decodeArrayOfFloats:(float *)result forKey:(NSString *)key {
   NSString *string=[self decodeObjectForKey:key];
   NSUInteger i,length=[string length],resultLength=0;
   unichar  buffer[length];
   float   multiplier=0.10f,sign=1,exponent=0,expsign=1;
   enum {
    expectingBraceOrSpace,
    expectingBraceSpaceOrInteger,
    expectingSpaceOrInteger,
    expectingInteger,
    expectingFraction,
    expectingExponent,
    expectingCommaBraceOrSpace,
    expectingSpace
   } state=expectingBraceOrSpace;
  
   if(string==nil)
    return NSNotFound;
    
   [string getCharacters:buffer];
    
   for(i=0;i<length;i++){
    unichar code=buffer[i];
     
    switch(state){
     
     case expectingBraceOrSpace:
      if(code=='{')
       state=expectingBraceSpaceOrInteger;
      else if(code>' ')
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
      
     case expectingBraceSpaceOrInteger:
      if(code=='{'){
       state=expectingSpaceOrInteger;
       break;
      }
      // fallthru     
     case expectingSpaceOrInteger:
      if(code<=' ')
       break;
      // fallthru
     case expectingInteger:
      if(code=='-')
       sign=-1;
      else if(code>='0' && code<='9')
       result[resultLength]=result[resultLength]*10+(code-'0');
      else if(code=='.'){
       multiplier=0.10f;
       state=expectingFraction;
      }
      else if(code=='e' || code=='E'){
       state=expectingExponent;
       exponent=0;
      }
      else if(code==','){
       result[resultLength++]*=sign;
       sign=1;
       state=expectingSpaceOrInteger;
      }
      else if(code=='}'){
       result[resultLength++]*=sign;
       sign=1;
       state=expectingCommaBraceOrSpace;
      }
      else if(code<=' '){
       result[resultLength++]*=sign;
       sign=1;
       state=expectingCommaBraceOrSpace;
      }
      else
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
       
     case expectingFraction:
      if(code>='0' && code<='9'){
       result[resultLength]=result[resultLength]+multiplier*(code-'0');
       multiplier/=10;
      }
      else if(code=='e' || code=='E'){
       state=expectingExponent;
       exponent=0;
      }
      else if(code==','){
       result[resultLength++]*=sign;
       sign=1;
       state=expectingSpaceOrInteger;
      }
      else if(code=='}'){
       result[resultLength++]*=sign;
       sign=1;
       state=expectingCommaBraceOrSpace;
      }
      else
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
     
     case expectingExponent:
      if(code=='+')
       break;
      if(code=='-')
       expsign=-1;
      else if(code>='0' && code<='9')
       exponent=exponent*10+(code-'0');
      else if(code==','){
       result[resultLength++]*=sign*powf(10.0f,expsign*exponent);
       sign=expsign=1;
       exponent=0;
       state=expectingSpaceOrInteger;
      }
      else if(code=='}'){
       result[resultLength++]*=sign*powf(10.0f,expsign*exponent);
       sign=expsign=1;
       exponent=0;
       state=expectingCommaBraceOrSpace;
      }
      else
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
     
     case expectingCommaBraceOrSpace:
      if(code==',')
       state=expectingBraceSpaceOrInteger;
      else if(code=='}')
       state=expectingSpace;
      else if(code>=' ')
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
      
     case expectingSpace:
      if(code>=' ')
       [NSException raise:NSInvalidArgumentException format:@"Unable to parse geometry %@, state=%d, pos=%d, code=%C",string,state,i,code];
      break;
    }
   }
   
   return resultLength; 
}

-(NSPoint)decodePointForKey:(NSString *)key {
   float    array[4]={ 0,0,0,0 };
   NSUInteger length=[self decodeArrayOfFloats:array forKey:key];
   
   return NSMakePoint(array[0],array[1]);
}

-(NSSize)decodeSizeForKey:(NSString *)key {
   float     array[4]={ 0,0,0,0 };
   NSUInteger length=[self decodeArrayOfFloats:array forKey:key];
   
   return NSMakeSize(array[0],array[1]);
}

-(NSRect)decodeRectForKey:(NSString *)key {
   float    array[4]={ 0,0,0,0 };
   NSUInteger length=[self decodeArrayOfFloats:array forKey:key];
   
   return NSMakeRect(array[0],array[1],array[2],array[3]);
}

-_decodeObjectWithPropertyList:plist {
   unsigned typeID=[plist _cfTypeID];

   if(typeID==kNSCFTypeString || typeID==kNSCFTypeData)
    return plist;
   if(typeID==kNSCFTypeDictionary)
    return [self decodeObjectForUID:integerFromCFUID(plist)];
   if(typeID==kNSCFTypeArray){
    NSMutableArray *result=[NSMutableArray array];
    NSInteger       i,count=[plist count];
    
    for(i=0;i<count;i++){
     id sibling=[plist objectAtIndex:i];
     
     [result addObject:[self _decodeObjectWithPropertyList:sibling]];
    }
    
    return result;
   }
   if([plist isKindOfClass:[CFUID class]])
    return [self decodeObjectForUID:[plist integerValue]];
   
   [NSException raise:@"NSKeyedUnarchiverException" format:@"Unable to decode property list with class %@",[plist class]];
   return nil;
}

-decodeObjectForKey:(NSString *)key {
   id result;
      
   id plist=[[_plistStack lastObject] objectForKey:key];
   
   if(plist==nil)
    result=nil;
   else
    result=[self _decodeObjectWithPropertyList:plist];

   return result;
}

-(void)replaceObject:object withObject:replacement {
   NSMapEnumerator state=NSEnumerateMapTable(_uidToObject);
   void           *key,*value;
   
   while(NSNextMapEnumeratorPair(&state,&key,&value)){
    if(value==object){
    
     if([_delegate respondsToSelector:@selector(unarchiver:willReplaceObject:withObject:)])
      [_delegate unarchiver:self willReplaceObject:value withObject:replacement];
      
     NSMapInsert(_uidToObject,key,replacement);
     return;
    }
   }
}

-(void)finishDecoding {
}

-delegate {
   return _delegate;
}

-(void)setDelegate:delegate {
   _delegate=delegate;
}

+(void)setClass:(Class)class forClassName:(NSString *)className {
}

+(Class)classForClassName:(NSString *)className {
   return Nil;
}

-(void)setClass:(Class)class forClassName:(NSString *)className {
   [_nameToReplacementClass setObject:class forKey:className];
}

-(Class)classForClassName:(NSString *)className {
   return [_nameToReplacementClass objectForKey:className];
}

-(NSString*)description
{
	return [NSString stringWithFormat:@"%@ %@", [super description], [_plistStack lastObject]];
}
@end
